/************************************************************************\
* Project Von Neumann - unfe.c code file                                 *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Description : Universe to FE output functions                          *
\************************************************************************/ 

/* for NULL: */
#include <stdlib.h>

/* PVN header files */
#include "untypes.h"
#include "unfe.h"
#include "fe.h"

extern UN_Object* **Sectors;
extern unsigned int size_sectx, size_secty;

/***************************************************************************\
* Output to FE
\***************************************************************************/

/***************************************************************************\
* Function Name: UN_InRange
* Description  : Checks if an object is in the range box given.
* Arguments    : object - the object to check
*                x, y   - the top coordinates of the range box
*                w, h   - size of the range box
* Returns      : TRUE if in range
*                FALSE if not in range
\***************************************************************************/
UN_Bool UN_InRange(UN_Object* object, UN_Coord x, UN_Coord y, 
		   unsigned int w, unsigned int h)
     {
	if (object->x < x - (long)object->w)
	  return FALSE;
	
	if (object->y < (y - (long)object->h))
	  return FALSE;
	
	if (object->x > (x + w))
	  return FALSE;
	
	if (object->y > (y + h))
	  return FALSE;

	/* if we are not out of range, we must be in range */
	return TRUE;
     }

/***************************************************************************\
* Function Name: UN_DrawAroundObject
* Description  : Draw with an object as focus
* Arguments    : object - the focus object
* Returns      : void
\***************************************************************************/
void UN_DrawAroundObject(UN_Object* object)
     {
	UN_Coord x, y;
 	
	/* determines if we are nearing the left and top borders of the
	 * field. If so, stop following the objects
	 * Of course the right and bottom borders also need checking, 
	 * I haven't implemented this yet. */
	
	x = (object->x - (int)(size_sectx/2) < 0 ? 0 : object->x - (size_sectx/2));
	y = (object->y - (int)((size_secty - (64 * 256))/2) < 0 ?
	     0 : object->y - ((size_secty - (64 * 256))/2));
   	
	/* draw all visible objects around it */
	/* NOTE: the minus 4 * 256 is a correction factor for the output.
	 * the max y coordinate on the virtual output screen is 768
	 * This is 256 smaller than 1024, the max x coordinate.
	 * The mapping is currently 4 to 1, and the sectors are 4096 by
	 * 4096 UN_Coords large. 
	 * Of course this all needs to be automated, by determining the
	 * factor (in this case 4) on the fly. */
	UN_DrawVisibleObjects(x, y, size_sectx, size_secty - (64 * 256));
     }

/***************************************************************************\
* Function Name: UN_DrawVisibleObjects
* Description  : Draw all visible objects within certain coordinates on the
*                screen.
* Arguments    : x, y - top left of the area we want to draw all objects in
*                w, h - size of area we want to draw
* Returns      : void
\***************************************************************************/
void UN_DrawVisibleObjects(UN_Coord x, UN_Coord y, 
			   unsigned int w, unsigned int h)
     {
	UN_Sector topsect_x, topsect_y;
	
	/* figure out which sectors we have to draw */
	
	/* top left sector */
	topsect_x = x / size_sectx;
	topsect_y = y / size_secty;
	
	/* draw the 4 sectors in this area */
	UN_DrawSector(x, y, w, h, topsect_x, topsect_y);
	UN_DrawSector(x, y, w, h, topsect_x + 1, topsect_y);
	UN_DrawSector(x, y, w, h, topsect_x, topsect_y + 1);
	UN_DrawSector(x, y, w, h, topsect_x + 1, topsect_y + 1);
     }

/***************************************************************************\
* Function Name: UN_DrawSector
* Description  : Draw all objects within range in a sector.
* Arguments    : i, j - coordinates of the sector
*                x, y, w, h - draw only objects that have coordinates that
*                are in this box.
* Returns      : void 
\***************************************************************************/
void UN_DrawSector(UN_Coord x, UN_Coord y, 
		   unsigned int w, unsigned int h, 
		   UN_Sector i, UN_Sector j)
     {
	UN_Object* object;
       
	/* go through all objects in sector */
	for (object = Sectors[i][j]->next_sect; /* skip dummy */
	     object != NULL; /* draw until we reached the end */
	     object = object->next_sect)
	  {
	     /* check if object is in range */
	     if (UN_InRange(object, x, y, w, h))
	       {
		  /* it is in range, so draw it */
		  FE_PushBitmap((object->x - x)/64,
				(object->y - y)/64,
				(object->w)/64,
				(object->h)/64,
				object->facedir,
				object->pict); 
	       }
	  }
     }
