
/************************************************************************\
* Project Von Neumann - ungame.c code file                               *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Note        : The physics of the game                                  *
\************************************************************************/ 

/* #define DEBUG for debugging text */
/* #define DEBUG */


#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/* for NULL */
#include <stdlib.h>

/* PVN header files */
#include "fe.h"
#include "untypes.h"
#include "unutil.h"
#include "ungame.h"
#include "unfe.h"

/* global variables */
UN_Object* Player; /* pointer to the player object */
extern UN_Object* ActiveObjects;
extern float UN_Sin[NUM_DIRS];
extern float UN_Cos[NUM_DIRS];
extern unsigned int ThrustCost[MAX_THRUST];
extern UN_Coord size_x, size_y;
extern unsigned int size_sectx, size_secty;

/****************************************************************************\
* Game physics functions.
\****************************************************************************/

/****************************************************************************\
* Function Name: UN_GameLoop
* Description  : The main game loop
* Arguments    : none
* Returns      : void
\****************************************************************************/
void UN_GameLoop()
     {
	while (1)
	  { 
	     UN_UpdateAll(); /* update all objects */
	     /* draw all visible objects around player to mem buffer */
	     UN_DrawAroundObject(Player);
	     /* UN_DrawSector(0, 0, 65565, 65556 - (256 * 64), 0, 0); */
	     
	     FE_UpdateScreen();  /* blit buffer to screen, update buffer */
	     FE_GetEsc();  /* check if we want to quit */
	  } 
      }

/****************************************************************************\
* Function Name: UN_UpdateAll
* Description  : Updates all objects in the universe.
* Arguments    : none
* Returns      : void
\****************************************************************************/
void UN_UpdateAll()
     {
	UN_Object* current;
	
	/* fetch player actions */
	Player->extra.animate.actions = FE_GetActions(); 
	
	/* fetch probe etc actions here eventually with 
	 * AL functions. Note that these AL functions are not going
	 * to do a lot of processing; Alife will do processing in its own
	 * time segment, semi-independent of universe. */
	
	/* go through active object list, and update all objects in it
	 * Note that active objects list should not be empty here */
	for (current = ActiveObjects->next_glob; /* skip dummy */ 
	     current != NULL; /* as long as we are not at the end */
	     current = current->next_glob) /* next object */
	  {
	     UN_UpdateObject(current);
	  }
     }

/****************************************************************************\
* Function Name: UN_UpdateObject
* Description  : Updates everything in an object, physics, collision checking
*                and actions if the object was animate.
* Arguments    : The object to update.
* Returns      : void
\****************************************************************************/
void UN_UpdateObject(UN_Object* object)
     {
	/* If the object is animate, apply its actions */
	if (object->what <= OBJ_MISSILE)
	  {
	     UN_ApplyActions(object);
#ifdef DEBUG
	     printf("Actions:%d velx:%f vely:%f face:%d\n",
		    ANIMATE.actions,
		    object->velx, object->vely, object->facedir);
#endif
	  }
	
	UN_ApplyPhysics(object); /* Do physics */
	/* UN_ApplyCollision(object); */ /* check for collisions */
     }

/***************************************************************************\ 
* Function Name: UN_ApplyActions
* Description  : Applies the actions as set in the animate object 
*                (in 'extra.animate.actions'), to that object.
* Arguments    : object - the object we apply the actions to
* Returns      : void
* Notes        : This function assumes the actions (keypresses) are 
*                set in a previous function.
*                The macro ANIMATE stands for object->extra.animate
*                The thrust and brake code currently uses some Allegro
*                fixed point routines (in an inefficient way, too. Just to
*                                      make it work)
\***************************************************************************/
void UN_ApplyActions(UN_Object* object)
     {
	static unsigned long curthrust;  /* temporary holder for thrust */
	
	if (ANIMATE.actions & ACT_THRUST) /* forward thrust */
	  {
	     curthrust = ANIMATE.thrust; /* fetch thrust from object */
	     if (object->energy > ThrustCost[curthrust]) /* if enough energy */
	       {
		  object->velx += 
		  ((float)curthrust / (float)object->mass) *
		  UN_Sin[object->facedir];
#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Sin:%f\n",
			 curthrust, object->mass, UN_Sin[object->facedir]);
#endif /* DEBUG */
		  object->vely -= 
		  ((float)curthrust / (float)object->mass) *
		  UN_Cos[object->facedir];
#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Cos:%f\n",
			 curthrust, object->mass, UN_Cos[object->facedir]);
#endif /* DEBUG */
		  
		  /* update energy */
		  object->energy -= ThrustCost[curthrust];
	       }
	     else /* not enough energy for thrust */
	       {
		  /* Need to set AS special effect for 
		   * failure of thrust (sound?) */
		  
		  /* hardware interrupt for this object.
		   * Nothing extensive happens here, just preparation for
		   * the Alife handler that's called later. */
		  /* AL_Interrupt(object, THRUST_FAIL); */
	       }
	  } /* end of THRUST code */
	else if (ANIMATE.actions & ACT_BRAKE) /* we can't thrust and brake at the same time, hence the else if */
	  {
	     curthrust = ANIMATE.thrust;
	     if (object->energy > ThrustCost[curthrust]) /* if enough energy */
	       {
		  object->velx -= 
		  ((float)curthrust / (float)object->mass) *
		  UN_Sin[object->facedir];

#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Sin:%f\n",
			 curthrust, object->mass, UN_Sin[object->facedir]);
#endif /* DEBUG */

		  object->vely += 
		  ((float)curthrust / (float)object->mass) *
		  UN_Cos[object->facedir];

#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Cos:%f\n",
			 curthrust, object->mass, UN_Cos[object->facedir]);
#endif /* DEBUG */

		  /* update energy */
		  object->energy -= ThrustCost[curthrust];
	       }
	     else /* not enough energy for brake thrust */
	       {
		  /* Need to set AS special effect for 
		     failure of brake (sound?) */
		  
		  /* hardware interrupt for this object.
		   * Nothing extensive happens here, just preparation for
		   * the Alife handler that's called later. */
		  /* AL_Interrupt(object, BRAKE_FAIL); */
	       } 
	  } /* end of BRAKE code - end of THRUST/BRAKE code */

	if (ANIMATE.actions & ACT_LEFT) /* turn left */
	  {
	     /* counter clockwise */
	     /* faster version, but less portable:
	      object->facedir = (object->facedir - ANIMATE.turn) & NUM_DIRS-1;
	      */
	     object->facedir = 
	     ((object->facedir > ANIMATE.turn)
	      ? (object->facedir - ANIMATE.turn)
	      : (NUM_DIRS - (ANIMATE.turn - object->facedir)));
	  } /* end of LEFT code */
	else if (ANIMATE.actions & ACT_RIGHT) /* turn right */
	  {
	     /* clockwise */
	     object->facedir = (object->facedir + ANIMATE.turn) & (NUM_DIRS-1);
	  } /* end of RIGHT code */
	
	/* target cycling for player: not yet */
	/* fire code: not yet */
     }
	     
/**************************************************************************\
* Function Name: UN_ApplyPhysics
* Description  : Applies physics to an object. This means an object
*                keeps moving with the same velocity.
* Arguments    : object - what we apply the physics to
* Returns      : void
* Notes        : Collision checking is done in UN_ApplyCollisions
\**************************************************************************/
void UN_ApplyPhysics(UN_Object* object)
     {
	/* Flag, set if we moved out of sector */
	UN_Bool moved_out;
	
	/* Update the age of the object */
	if ((object->what == OBJ_BULLET) ||
	    (object->what == OBJ_MISSILE))
	  {
	     /* bullets and missiles have a maximum lifespan that is
	      * set at their creation. Their age decreases until it
	      * is 0. When the age is 0, their shield is set to 0,
	      * meaning the object will now die. */
	     (object->age)--;
	     if (object->age == 0)
	       object->shield = 0;
	  }
	else
	  {
	     /* other objects have increasing ages */
	     (object->age)++; 
	  }
	
	/* update the x and y of the object */
	
	object->x += object->velx;
	object->y += object->vely;
	
	/* it possible we now moved out of the playing field 
	 * For ease of coding, UN_Coord is *signed*. 
	 * This means the max field is 2 to the power 31 large, assuming
	 * 32 bit longs. This is about 2,147,000,000 maximum.  
	 * Large enough. */
	
	/* NOTE: currently, we have 'stone wall' borders 
	 * This because allowing wrap makes things like range
	 * checking (used in the output) very tricky. */
	
	/* if we moved out on the left, move in from the right */
	/* if (object->x < 0)
	  object->x = size_x + object->x; */
	
	if (object->x < 0)
	  object->x = 0;
	
	/* if we moved out on the top, move in from the bottom */
	/* if (object->y < 0)
	  object->y = size_y + object->y; */
	
	if (object->y < 0)
	  object->y = 0;
	
	/* if we moved out on the right, move in from the left */
	/* if (object->x >= size_x)
	  object->x = object->x - size_x; */
	
	if (object->x >= size_x - object->w)
	  object->x = size_x - object->w;
	
	/* if we moved out on the bottom, move in from the top */
	/* if (object->y >= size_y)
	  object->y = object->y - size_y; */
	
	if (object->y >= size_y - object->h)
	  object->y = size_y - object->h;
	
	/* now check if we moved out of our sector */
	moved_out = FALSE;
	if (object->x / size_sectx != object->sectx)
	  {
	     object->sectx = object->x / size_sectx;
	     moved_out = TRUE;
	  }
	
	if (object->y / size_secty != object->secty)
	  {
	     object->secty = object->y / size_secty;
	     moved_out = TRUE;
	  }
	
	/* if we indeed moved out of our sector, update object lists */
	if (moved_out)
	  UN_ShiftObject(object);               
     }

void UN_ApplyCollisions(UN_Object* object)
     {
	/* No collision detection yet, until the steering part works */
     }

