/************************************************************************\
* Project Von Neumann - uninit.c code file                               *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Description : Universe initialisation functions                        *
\************************************************************************/

/* Note that the final universe version would not need to refer to
 * Allegro or SVGALIB at all. The code that conditially compiles
 * is needed to initialize the objects to a default bitmap (ship)
 */

/* for malloc */
#include <stdlib.h>

#include "fever.h"

#ifdef ALLEGRO
#include <allegro.h> /* for bitmap definition */
#endif 

/* PVN header files */
#include "untypes.h"
#include "uninit.h"
#include "unutil.h"

/***************************************************************************\
* Description: Global variables
\***************************************************************************/
/* temporary: ship picture */
#ifdef ALLEGRO
extern BITMAP* ship;
#endif /* ALLEGRO */
#ifdef SVGALIB
extern unsigned char ship[32 * 32];
#endif /* SVGALIB */ 
#ifdef TEXTOUT
extern unsigned char ship[32 * 32];
#endif /* TEXTOUT */
 
/* player object */
extern UN_Object* Player;

UN_Object* **Sectors; /* dynamically allocated 2-d array of pointers to object lists */
UN_Object* ActiveObjects; /* active object list */
UN_Object* SpareStack = NULL; /* spare object stack */

/* trig lookup tables */
float UN_Sin[NUM_DIRS];
float UN_Cos[NUM_DIRS];

/* some parameter like global variables */
unsigned int ThrustCost[MAX_THRUST];
UN_Coord size_x, size_y;
UN_Coord size_sectx, size_secty;

/****************************************************************************\
* Initialisation functions.
\****************************************************************************/

/*************************************************************************\
* Function Name: UN_InitSectors
* Description  : Create a 2 dimensional array of sectors (pointers to objects)
* Arguments    : x_sectors, y_sectors - width and height of the 2-d array  
* Returns      : void
* Notes        : the sector is filled with dummy objects.
\*************************************************************************/
void UN_InitSectors(unsigned int x_sectors, unsigned int y_sectors)
     {
	unsigned int i, j;
	UN_Object* dummy;
	
	Sectors = (UN_Object* **) malloc(sizeof(UN_Object*) * x_sectors);
	
	for (i = 0; i < x_sectors; i++)
	  Sectors[i] = (UN_Object* *) malloc(sizeof(UN_Object*) * y_sectors);

	/* NOTE: we might need to check if malloc returns NULL */
	
	/* the sector is filled with dummy objects */
	for (i = 0; i < x_sectors; i++)
	  {       
	    for (j = 0; j < y_sectors; j++)
	       {
		 /* create new dummy object */
		  dummy = UN_NewObject();
		  /* initialize it */
		  dummy->sectx = i;
		  dummy->secty = j;
		  dummy->prev_sect = NULL;
		  dummy->next_sect = NULL;
		  /* let the sectors list point at it */
		  Sectors[i][j] = dummy;
	       }
	  }
	
     }

/****************************************************************************\
* Function Name: UN_InitObj
* Description  : Initializes an object
* Arguments    : object - which object
* Returns      : void
\****************************************************************************/
void UN_InitObj(UN_Object* object)
     {
	object->age = 0;
	object->next_sect = NULL;
	object->prev_sect = NULL;
	object->next_glob = NULL;
	object->prev_glob = NULL;
     }

/****************************************************************************\
* Function Name: UN_InitObjCoord
* Description  : Initializes the location of an object, size, 
*                and the way it faces
* Arguments    : object - which object
*                x, y - coordinates
*                w, h - size
*                facedir - way it faces
* Returns      : void
\****************************************************************************/
void UN_InitObjCoord(UN_Object* object, UN_Coord x, UN_Coord y,
		  unsigned int w, unsigned int h,
		  UN_Direction facedir)
     {
	object->x = x;
	object->y = y;
	object->sectx = x / size_sectx;
	object->secty = y / size_secty;
	object->w = w;
	object->h = h;
	object->facedir = facedir;
     }

/****************************************************************************\
* Function Name: UN_InitObjVelocity
* Description  : Initializes the velocity of an object
* Arguments    : object - which object
*                velx, vely - velocity vector
* Returns      : void
\****************************************************************************/
void UN_InitObjVelocity(UN_Object* object, UN_Velocity velx, UN_Velocity vely)
     {
	object->velx = velx;
	object->vely = vely;
     }

/****************************************************************************\
* Function Name: UN_InitObjType
* Description  : Initializes the type of an object
* Arguments    : object - which object
*                type   - what type
* Returns      : void
\****************************************************************************/
void UN_InitObjType(UN_Object* object, UN_ObjType type)
     {
	object->what = type;
     }

/****************************************************************************\
* Function Name: UN_InitObjStats
* Description  : Initializes the starting stats of an object.
* Arguments    : object - which object
*                shield - amount of shield
*                energy - amount of energy
*                mass   - mass of this object
* Returns      : void
\****************************************************************************/
void UN_InitObjStats(UN_Object* object, long shield, long energy, long mass)
     {
	object->shield = shield;
	object->energy = energy;
	object->mass = mass;
     }

/**************************************************************************\
* Function Name: UN_InitObjAnimate
* Description  : Initializes the animate part of an object
* Arguments    : turn - how fast this object turns
*                thrust - how much thrust power this object has
*                actions - initial action bitstring of this object
\**************************************************************************/
void UN_InitObjAnimate(UN_Object* object, unsigned char turn, 
		       unsigned long thrust, 
		       UN_Actions actions)
     {
	ANIMATE.turn = turn; 
	ANIMATE.thrust = thrust;
	ANIMATE.actions = actions;
     }

/****************************************************************************\
* Function Name: UN_InitThrust
* Description  : Initializes the ThrustCost lookup table
* Arguments    : none
* Returns      : void
\****************************************************************************/
void UN_InitThrust()
     {
	int i;
	
	for (i = 0; i < MAX_THRUST; i++)
	  ThrustCost[i] = i;
     }

/****************************************************************************\
* Function Name: UN_Init
* Description  : Initializes structures and variables needed by universe.
* Arguments    : none
* Returns      : void
\****************************************************************************/
void UN_Init()
     {
	/* trig is currently initialized in FE_Init().
	 * We may want to move this back to universe. */
	UN_Object* other1;
	UN_Object* other2;
	UN_Object* other3;
	
	/* initialisation of trig functions is currently done in the front
	 * end. We may want to move it back to universe */
	
	/* temporary size, these should be arguments of something */
	size_x      = 65536 * 16;
	size_y      = 65536 * 16;
	size_sectx  = 65536;
	size_secty  = 65536;
	
	/* initialize sectors */
	UN_InitSectors(size_x / size_sectx, size_y / size_secty);
	
	/* initialize active objects list */
	ActiveObjects = UN_NewObject();
	ActiveObjects->next_glob = NULL;
	ActiveObjects->prev_glob = NULL;
	
	/* initialize thrust cost lookup table */
	UN_InitThrust(); 
	
	/* make the player */
	Player = UN_NewObject();
	/* set player picture */
	Player->pict = ship;
	
	/* -- Test initialisation follows -- */
	
	/* initialize the player */
	UN_InitObj(Player);
	UN_InitObjType(Player, OBJ_PLAYER);
	UN_InitObjCoord(Player, 1000, 1000, 32 * 64, 32 * 64, 0);
	UN_InitObjVelocity(Player, 0, 0);
	UN_InitObjStats(Player, 100, 100000, 10);
	UN_InitObjAnimate(Player, 1, 10, 0);
		
	/* make a few other objects */
	other1 = UN_NewObject();
	other2 = UN_NewObject();
	other3 = UN_NewObject();
	
	/* they have a  picture */
	other1->pict = ship;
	other2->pict = ship;
	other3->pict = ship;
	
	UN_InitObj(other1);
	UN_InitObjType(other1, OBJ_PROBE);
	UN_InitObjCoord(other1, 10000, 10000, 32 * 64, 32 * 64, 0);
	UN_InitObjVelocity(other1, 10, 0);
	UN_InitObjStats(other1, 100, 1000, 1);
	UN_InitObjAnimate(other1, 1, 1, 0); UN_InitObj(other1);
	
	UN_InitObjType(other2, OBJ_PROBE);
	UN_InitObjCoord(other2, 20000, 10000, 32 * 64, 32 * 64, 30);
	UN_InitObjVelocity(other2, 0, 100);
	UN_InitObjStats(other2, 100, 1000, 1);
	UN_InitObjAnimate(other2, 1, 1, 0);
	
	UN_InitObj(other3);
	UN_InitObjType(other3, OBJ_PROBE);
	UN_InitObjCoord(other3, 10000, 20000, 32 * 64, 32 * 64, 200);
	UN_InitObjVelocity(other3, 30, 50);
	UN_InitObjStats(other3, 100, 1000, 1);
	UN_InitObjAnimate(other3, 1, 1, 0);
	
	/* activate everything */
	UN_ActivateObject(Player);
	UN_ActivateObject(other1);
	UN_ActivateObject(other2);
	UN_ActivateObject(other3);
	
	/* put everything in their own sectors */
	UN_ToSector(Player);
	UN_ToSector(other1);
	UN_ToSector(other2);
	UN_ToSector(other3);
     }   

