/************************************************************************\
* Project Von Neumann - unutil.c code file                               *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Description : Utility functions used by universe                       *
\************************************************************************/ 

/* if DEBUG is defined, some functions will emit text. Do not #define
* DEBUG unless the front end is in TEXTOUT mode! 
* (although SVGALIB does indeed seem to work with this) */
/* #define DEBUG */

/************************************************************************\
* Description: header files
\************************************************************************/

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/* Standard header files */
#include <stdlib.h> /* mainly for malloc */

/* PVN header files */
#include "unutil.h"
#include "untypes.h"

/***************************************************************************\
* Description: Global variables
\***************************************************************************/
UN_Object* **Sectors; /* dynamically 2-d array of pointers to object lists */
UN_Object* ActiveObjects; /* active object list */
UN_Object* SpareStack; /* spare object stack */

/*************************************************************************\
* Utility functions
\*************************************************************************/
  
/*************************************************************************\ 
* Function Name: UN_NewObject
* Description  : Obtain a new UN_Object
* Arguments    : none  
* Returns      : pointer to newly allocated UN_Object
\*************************************************************************/
UN_Object* UN_NewObject()
     {
	UN_Object* object;
	
	if (SpareStack != NULL)
	  {
	     /* SpareStack is not empty */
	     object = SpareStack;
	     SpareStack = SpareStack->next_sect;
	     return (UN_Object*) object;
	  }
	else
	  {
	     /* SpareStack is empty, allocate a new object */
	     object = (UN_Object*) malloc(sizeof(UN_Object));
	     /* NOTE: we want to check for NULL being returned.
	      * (as then no allocation took place) */
	     return (UN_Object*) object;
	  }
     }

/***************************************************************************\
* Function Name: UN_DeleteObject
* Description  : Delete an object
* Arguments    : object - the object to delete
* Returns      : void
* Notes        : this assumes all object lists start with a dummy object.
\***************************************************************************/
void UN_DeleteObject(UN_Object* object)
     {
	/* Remove object from the sector list it was in */
	
	/* First remove the object from the old list */
	if (object->next_sect != NULL) /* if we are not last object in list */
	  {
	     /* connect next_sect object to prev_sect object */
	     object->next_sect->prev_sect = object->prev_sect; 
	  }
	/* connect prev_sect object to next_sect object */
	object->prev_sect->next_sect = object->next_sect;
	
	/* remove the object from the active object list it was in */
        if (object->next_glob != NULL) /* if we are not last object in list */
	  {
	     /* connect next_glob object to prev_glob object */
	     object->next_glob->prev_glob = object->prev_glob;
	  }
	/* connect prev_glob object to next_glob object */
	object->prev_glob->next_glob = object->next_glob;
	 
	/* Proceed to put object on the top of the spare stack */
	object->next_sect = SpareStack;
	SpareStack = object;
     }

/***************************************************************************\
* Function Name: UN_ToSector(UN_Object* object)
* Description  : Links an object into a sector list.
* Arguments    : object - the object
* Returns      : void
\***************************************************************************/
void UN_ToSector(UN_Object* object)
     {
	/* we do not have to remove the object from anything */
	
	/* Insert the object at the top of the new sector list,
	 * under the dummy (which is always the first object) */
	
	/* connect this object to start of rest of list */
	object->next_sect = Sectors[object->sectx][object->secty]->next_sect;
	if (object->next_sect != NULL) /* if we are not last object in list */
	  {
	     /* connect next_object to this object */
	     object->next_sect->prev_sect = object;
	  }
	/* connect this object to start of list (the dummy) */
	object->prev_sect = Sectors[object->sectx][object->secty];
	/* connect dummy to this object */
	Sectors[object->sectx][object->secty]->next_sect = object;
	
#ifdef DEBUG
	printf("UN_ToSector: sectx:%lu secty:%lu\n", 
	       object->sectx, object->secty);
#endif /* DEBUG */
     }
/***************************************************************************\
* Function Name: UN_ShiftObject
* Description  : Removes the object from the sector object list it is in, and 
*                then puts it in the object list of a new sector.
* Arguments    : object - the object to move
* Returns      : void
\***************************************************************************/
void UN_ShiftObject(UN_Object* object)
     {
	/* First remove the object from the old list */
	
	if (object->next_sect != NULL) /* if we are not last object in list */
	  {
	     /* connect next_sect object to prev_sect object */
	     object->next_sect->prev_sect = object->prev_sect;  
	  }
	/* connect prev_sect object to next_sect object */
	object->prev_sect->next_sect = object->next_sect; 
	
	/* Then insert the object at the top of the new sector list,
	 * under the dummy (which is always the first object) */
	
	/* connect this object to start of rest of list */
	object->next_sect = Sectors[object->sectx][object->secty]->next_sect;
	if (object->next_sect != NULL) /* if we are not last object in list */
	  {
	     /* connect next_object to this object */
	     object->next_sect->prev_sect = object;
	  }
	/* connect this object to start of list (the dummy) */
	object->prev_sect = Sectors[object->sectx][object->secty];
	/* connect dummy to this object */
	Sectors[object->sectx][object->secty]->next_sect = object;
	
#ifdef DEBUG
	printf("UN_ShiftObject: sectx:%lu secty:%lu\n", 
	       object->sectx, object->secty);
#endif /* DEBUG */
     }

/***************************************************************************\
* Function Name: UN_ActivateObject
* Description  : Adds an object to top of the ActiveObjects list 
*                (the active objects list) 
* Arguments    : the object to activate
* Returns      : void
* Notes        : Assumes ActiveObjects list has a dummy object on top.
\***************************************************************************/
void UN_ActivateObject(UN_Object* object)
     {
	/* connect next_glob of this object to previous top of the list */
	object->next_glob = ActiveObjects->next_glob;
	if (object->next_glob != NULL) /* if we are not last in list */
	  {
	     /* connect next object to this object */
	     object->next_glob->prev_glob = object;
	  }
	
	/* connect this object to the dummy object */
	object->prev_glob = ActiveObjects;
	ActiveObjects->next_glob = object;
     }

/**************************************************************************\
* Function Name: UN_DeActivateObject
* Description  : Removes the object from the ActiveObjects list
* Arguments    : the object to deactivate
* Returns      : void
* Notes        : Assumes ActiveObjects list has a dummy object on top.
\**************************************************************************/
void UN_DeActivateObject(UN_Object* object)
     {
	/* connect previous object to next object */
	object->prev_glob->next_glob = object->next_glob;
	if (object->next_glob != NULL) /* if we are not last object in list */
	  {
	     /* connect next object to previous object */
	     object->next_glob->prev_glob = object->prev_glob;
	  }
     }

